module compute
	use globals
	implicit none
	
	contains
	
    subroutine prep
        integer :: i
        integer, parameter  :: nimp0=30
        real    :: gqalt(nalt,2)
        
        
        call mkGQxw(gqalt)
        altgrid=xsimin+(xsimax-xsimin)*gqalt(:,1)
        walt0=gqalt(:,2)
        
        impgrid=[(xsimin,i=1,5),(xsimin+(xsimax-xsimin)*(i/real(nimp0-1)),i=0,nimp0-1),(xsimax,i=1,5)]
        nimp=size(impgrid)
        
        chkgrid=[(xsimin+(xsimax-xsimin)*(i/real(nchk-1)),i=0,nchk-1)]
        appgrid=[(xsimin+(xsimax-xsimin)*(i/real(napp-1)),i=0,napp-1)]
		if(allocated(Zs)) then
			deallocate(Zs)
			allocate(Zs(k+1,nsim))
		else
	        allocate(Zs(k+1,nsim),densaltraw(nalt,nsim), densapp(napp,nsim), densaltz(nalt,nsim),genLRstats(nsim))
		endif
        
    end subroutine
    
   function getqxsi(xsi) result(val)
        real    :: xsi, val
        if(quantileflag) then
            val=(htarget**(-xsi)-1)/xsi
        else
            val=(htarget**(-xsi)/(1-xsi)-1)/xsi
        endif
   end function
   
   elemental function log1p(x) result(val)
        real, intent(in)    :: x
		real	:: val
        real    :: z
        z=1+x
        val=log(z)-((z-1)-x)/z
   end function

    function logfsz(s,z,xsi) result (val)
        real    :: s,z(k+1),xsi,val
        real    :: xs(k), ys, q
        
        xs=z(1:k)
        ys=z(k+1)
        q=getqxsi(xsi)
        val=(k-1)*log(abs(q-s))-k*log(abs(ys))-exp(-log1p(xsi*s)/xsi)-(1+1/xsi)*sum(log(1+xsi*(s+xs*(q-s)/ys)))
    end function
   
    function dlogfsz(s,z,xsi) result (val)
        real    :: s,z(k+1),xsi,val
        real    :: xs(k), ys, q,qms
        
        xs=z(1:k)
        ys=z(k+1)
        qms=getqxsi(xsi)-s
        val=exp(-log1p(xsi*s)/xsi)-(1+1/xsi)*sum(xsi*(1-xs/ys)/(1+xsi*(s+xs*qms/ys)))-(k-1)/qms
    end function

    function logfz(z,xsi) result(val)
        use qdagi_int
        use qdag_int
        use IERCD_int
        real, parameter :: eps=0.00001
		real			:: gtol=1E-2

		real		:: val,z(k+1),xsi,qxsi,ys
        real        :: a,b,c
        globalz=z
        globalxsi=xsi
        intfunc=>logfsz
        dintfunc=>dlogfsz
        ys=z(k+1)
        qxsi=getqxsi(xsi)
        
        a=-huge(1.0);b=huge(1.0)
        c=-1/xsi
        if(c<0) then 
            a=max(a,c)
        else
            b=min(b,c)
        endif
        c=(ys+qxsi*xsi)/(xsi*(1-ys))
        if(xsi-xsi/ys>0) then 
            a=max(a,c)
        else
            b=min(b,c)
        endif
        if(ys>0) then
            b=min(b,qxsi)
        else
            a=max(a,qxsi)
		endif
		c=(-1 + (1000.0)**(-xsi/(1 + xsi)))/xsi
		if(c<0) a=max(a,c)
		if(b<a) then
			val=-1000
			return
!$omp critical					  
					 call disp(z)
                     call disp([xsi,a,b,qxsi])
                     call savevec("c:/dropbox/mystuff/tail/text.txt",z)
					 stop
                     printflag=.true.
!$omp end critical					  
		endif
        val=igab(a,b)

    contains

        function igab(a,b) result(val)
            real    :: a,b,val
            if(a==-huge(1.0)) then
                val=iinfb(b)
                return
            endif
            if(b==huge(1.0)) then 
                val=iainf(a)
            else
                val=iab(a,b)
            endif
        end function
        
        function iab(a,b) result (val)
            use uvmid_int
            real    :: val,a,b,m,val0
            real    :: xguess
            
            if(a*b<0) then
				if(abs(a)<abs(b)) then
                    xguess=a+min(1.0,.5*(b-a))
                else
                    xguess=b-min(1.0,.5*(b-a))
				endif				
            else 
                if(a>0) then 
                    xguess=a+min(1.0,.5*(b-a))
                else
                    xguess=b-min(1.0,.5*(b-a))
                endif
			endif
			if(b-a<.1) then
				m=.5*(a+b)
			else
				call uvmid(ext_minuslogf,ext_minusdlogf,a+eps,b-eps,m,xguess=xguess,gtol=gtol)
			endif
            globalls=logfsz(m,z,xsi)
            call qdag(ext_f,a,m,val0,ERRREL=quaderr,ERRABS=0.0)
            call qdag(ext_f,m,b,val,ERRREL=quaderr,ERRABS=0.0)
            val=log(val+val0)+globalls
        end function
        
        function iainf(a) result (val)
            use uvmid_int
            real    :: val,a,m,val0
            real    :: xguess
            
            if(dlogfsz(a+eps,z,xsi)<0) then
                globalls=logfsz(a+eps,z,xsi)
                call qdagi(ext_f,a,1,val,ERRREL=quaderr,ERRABS=0.0)
                val=log(val)+globalls
            else
                xguess=a+1
                call uvmid(ext_minuslogf,ext_minusdlogf,a+eps,a+1E4,m,xguess=xguess,gtol=gtol)
                globalls=logfsz(m,z,xsi)
                call qdag(ext_f,a,m,val0,ERRREL=quaderr,ERRABS=0.0)
                call qdagi(ext_f,m,1,val,ERRREL=quaderr,ERRABS=0.0)

                val=log(val+val0)+globalls
            endif
        end function

        function iinfb(b) result (val)
            use uvmid_int
            real    :: val,b,m,val0
            real    :: xguess
            
            if(dlogfsz(b-eps,z,xsi)>0) then
                globalls=logfsz(b-eps,z,xsi)
                call qdagi(ext_f,b,-1,val,ERRREL=quaderr,ERRABS=0.0)
                val=log(val)+globalls
            else
                xguess=b-1
                call uvmid(ext_minuslogf,ext_minusdlogf,b-1E4,b-eps,m,xguess=xguess,gtol=gtol)
                globalls=logfsz(m,z,xsi)
                call qdag(ext_f,m,b,val0,ERRREL=quaderr,ERRABS=0.0)
                call qdagi(ext_f,m,-1,val,ERRREL=quaderr,ERRABS=0.0)
                val=log(val+val0)+globalls
            endif
        end function
    end function

    function logfsx(s,z,xsi) result (val)
        real    :: s,z(k+1),xsi,val
        real    :: xs(k)
        
        xs=z(1:k)
        val=(k-1)*log(abs(xsi*s))-(1+1/xsi)*sum(log1p(xsi*s*xs))
    end function
   
    function dlogfsx(s,z,xsi) result (val)
        real    :: s,z(k+1),xsi,val
        real    :: xs(k)
        
        xs=z(1:k)
        val=(k-1)/s-(1+1/xsi)*sum(xsi*xs/(1+xsi*s*xs))
    end function

    function logfgxs(z,xsi) result(val)
        use qdagi_int
        use qdag_int
        use uvmid_int
        use IERCD_int
        real, parameter :: eps=0.01 
		real			:: gtol=1E-2

		real		:: val,z(k+1),xsi
        real        :: b,xguess,m,val0
        
        globalz=z
        globalxsi=xsi
        intfunc=>logfsx
        dintfunc=>dlogfsx
        b=merge(-1/xsi,1E4,xsi<0)
        xguess=1.0
        call uvmid(ext_minuslogf,ext_minusdlogf,eps,b-eps,m,xguess=xguess,gtol=gtol)
        globalls=logfsx(m,z,xsi)
        call qdag(ext_f,0.0,m,val0,ERRREL=quaderr,ERRABS=0.0)
        if(xsi>0) then
            call qdagi(ext_f,m,1,val,ERRREL=quaderr,ERRABS=0.0)
        else
            call qdag(ext_f,m,b,val,ERRREL=quaderr,ERRABS=0.0)
        endif
        val=globalls+log(val+val0)+log_gamma(k-xsi)-(k-1)*log(abs(xsi))
    end function
        
    subroutine setZ(U,Z,xsi)
		real		:: U(k),Z(k+1), xsi
		integer		:: i
		real, parameter	:: may=0.0005
		
		if(xsi>0) then
			z(1)=(-log(U(1)))**(-xsi)
			do i=2,k
				z(i)=(-log(U(i))+z(i-1)**(-1/xsi))**(-xsi)
            enddo
			z(k+1)=(1+xsi*getqxsi(xsi)-z(k))/(z(1)-z(k))
		else
			z(1)=-(-log(U(1)))**(-xsi)
			do i=2,k
				z(i)=-(-log(U(i))+(-z(i-1))**(-1/xsi))**(-xsi)
			enddo
			z(k+1)=(-(1+xsi*getqxsi(xsi))-z(k))/(z(1)-z(k))
		endif
		z(1:k)=(z(1:k)-z(k))/(z(1)-z(k))
		if(abs(z(k+1))<may) then
			z(k+1)=sign(may,z(k+1))	
		endif
    end subroutine
    
   	subroutine setdens()
		integer		:: i,l
		real		:: Us(k,nsim),impevals(nimp),z(k+1)
		
		do i=1,nsim
			call rnun(Us(:,i))
        enddo
        
        globalls=00
        
!$omp parallel do private (impevals,z,i,l)
		do gl=1,nsim
			call erset(0,1,0)
			call erset(1,0,0)
			call erset(2,0,0)
			call erset(3,0,0)
			call erset(4,1,0)
			call erset(5,1,0)
			l=gl
            call setZ(Us(:,l),z,impgrid(mod(l,nimp)+1))
!			call disp(z)
            do i=1,nimp
                impevals(i)=logfz(z,impgrid(i))
            enddo
            ldensimp(l)=logmeanexp(impevals)
            do i=1,nalt
                densaltraw(i,l)=exp(logfgxs(z,altgrid(i))-ldensimp(l))
                densaltz(i,l)=exp(logfz(z,altgrid(i))-ldensimp(l))
			enddo
            do i=1,napp
                densapp(i,l)=exp(logfz(z,appgrid(i))-ldensimp(l))
			enddo
            Zs(:,l)=Z
		enddo  
	end subroutine    
	
	function getcv(stats,w,alpha) result(val)
        use SVRGP_INT
        real    :: stats(nsim), w(nsim),alpha, val
        real    :: sort(nsim),p
        integer :: inds(nsim),l
        
        inds=[(l,l=1,nsim)]
        call svrgp(stats,sort,inds)
        p=w(inds(nsim))
        do l=nsim-1,1,-1
            if(p+w(inds(l))>alpha*nsim) then
                val=sort(l+1)
                return
            endif
            p=p+w(inds(l))
        enddo
        print *,"error in getcv: weights sum up to less than alpha"
        stop
	end function
	
	function selxsitab result(val)
		integer, parameter	:: nxsi=5
		real	:: val(3,nxsi)
		real, parameter:: xsigrid(nxsi)=[-.5,-.25,.01,.25,.5]
		integer		:: i,l
		real, allocatable	:: densaltold(:)
		logical, allocatable :: opttest(:)

		densaltold=densalt
		opttest=matmul(lam,densapp)>densalt
		do i=1,nxsi
!$omp parallel do 
			do l=1,nsim
				densalt(l)=exp(logfgxs(zs(:,l),xsigrid(i))-ldensimp(l))
			enddo
			call setlam
			call adjustlam
			call chksize
			val(3,i)=sum(densalt,mask=matmul(lam,densapp)>densalt)/nsim
			val(1,i)=sum(densalt,mask=genLRstats<cvgenLR)/nsim
			val(2,i)=sum(densalt,mask=opttest)/nsim
		enddo
		densalt=densaltold
		print *,"lengths of gLR, opt and envelope"
		call mdisp(xsigrid.cud.val)
	end function
	
	subroutine setwalt
		integer	:: l,i
		real	:: cv,lengths(nalt)
		
		do i=1,nalt
			cv=getcv(densaltraw(i,:)/densaltz(i,:),densaltz(i,:),alpha)
			lengths(i)=sum(densaltraw(i,:),mask=cv*densaltz(i,:)>densaltraw(i,:))/nsim
		enddo
		walt=walt0/lengths
		densalt=matmul(walt,densaltraw)
	end subroutine
	
	subroutine setlam()
		integer, parameter	:: n=napp
		real	:: RP(n), lamstep
		real	:: oldRP(n), lamfac(n), cv
		integer	:: l,j

		lamfac=4
		lam=1
		oldRP=alpha
		do j=1,500
			RP=0.0
			do l=1,nsim
				if(dot_product(densapp(:,l),lam)<densalt(l)) RP=RP+densapp(:,l)
			enddo
			RP=RP/nsim
			lam=lam*exp(lamfac*(RP-alpha))
			where((RP-alpha)*(oldRP-alpha)>0)
				lamfac=lamfac*1.03
			elsewhere
				lamfac=lamfac/2.0
			endwhere
			oldRP=RP
			if(j>400) lamfac=0.1
			call setbounds(lamfac,0.01,20.0)
		enddo		
		call disp(appgrid.cvr.RP.cud.lam)
		print *,sum(lam*RP)/sum(lam)
	end subroutine
	
	subroutine adjustlam
		real	:: RP,cv,avdensapp(nsim),wlength0,wlengthc, lamstep
		integer	:: j,l
		
		cv=sum(lam)
		lam=lam/cv
		avdensapp=matmul(lam,densapp)
		
		lamstep=0.1
		do j=1,20
			RP=0.0
			do l=1,nsim
				if(cv*avdensapp(l)<densalt(l)) RP=RP+avdensapp(l)
			enddo
			RP=RP/nsim
			if(RP>alpha) then
				cv=cv*exp(lamstep)
			else
				cv=cv*exp(-lamstep)
			endif
			lamstep=lamstep/2
		enddo
		if(abs(RP-alpha)>0.001) print '(a,f12.4)','problem in cv determination: average RP',RP

		wlength0=(1+lengthfudge)*sum(densalt,mask=cv*avdensapp>densalt)/nsim
		print *,"Length of optimal interval", wlength0
		lamstep=0.5
		do j=1,20
			wlengthc=sum(densalt,mask=cv*avdensapp>densalt)/nsim
			if(wlength0>wlengthc) then
				cv=cv*exp(lamstep)
			else
				cv=cv*exp(-lamstep)
			endif
			lamstep=lamstep/2
		enddo
		if(abs(wlengthc/wlength0-1)>0.0005) then
			print *,"problem in length-correction--target and achieved:"
			call disp([wlength0,wlengthc])
		endif
		lam=lam*cv
	end subroutine
	
	subroutine chksize
		real	:: RP(nchk),RP2(nchk),RPc(nchk)
		integer	:: l,i
		
		RP=0;RP2=0
!$omp parallel do private(RPc,i) reduction(+:RP,RP2) 
		do l=1,nsim
			if(dot_product(lam,densapp(:,l))<densalt(l)) then
				do i=1,nchk
					RPc(i)=exp(logfz(zs(:,l),chkgrid(i))-ldensimp(l))
				enddo
				RP=RP+RPc
				RP2=RP2+RPc**2
			endif
		enddo
		RP=RP/nsim
		RP2=RP2/nsim
		RP2=sqrt((RP2-RP**2)/nsim)
		if(maxval(RP)>alpha) then 
			print *,"warning: Size isn't controlled", alpha, maxval(RP)
			call disp(chkgrid.cvr.RP.cud.RP2)
		endif
	end subroutine
		
end module